/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Doug Daniels <rainking@alumni.rice.edu>
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Matthew Wilson <matthew@mjwilson.demon.co.uk>.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Constants.
 */
const NS_REMOTE_ANNOTATIONS_SERVER_CID = Components.ID("{8242c0de-9b27-4ee6-aab8-0b7778233b54}");
const NS_REMOTE_ANNOTATIONS_SERVER_PROG_ID = "@mozilla.org/remote_annotations_server;1";


/**
 * Global vars
 */

// global RDF Service
var gRDFService = Components
    .classes["@mozilla.org/rdf/rdf-service;1"]
    .getService(Components.interfaces.nsIRDFService);


function RemoteServer() {
    this.mUrl = null;
    this.mUsername = null;
    this.mPassword = null;
    this.initialized = false;
}


RemoteServer.UPDATE       = 0;
RemoteServer.DELETE       = 1;
RemoteServer.GET_REPLIES  = 2;
RemoteServer.POST         = 3;
RemoteServer.LOAD         = 4;
RemoteServer.REPLY        = 5;
RemoteServer.LOAD_BODY    = 6;
RemoteServer.AUTO_LOAD    = 7;

RemoteServer.ANNOT_NS = "http://www.w3.org/2000/10/annotation-ns#";
RemoteServer.ANNOT_TYPE_NS = "http://www.w3.org/2000/10/annotationType#";
RemoteServer.RDF_NS   = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
RemoteServer.DC_NS    = "http://purl.org/dc/elements/1.0/";
RemoteServer.THREAD_NS= "http://www.w3.org/2001/03/thread#";
RemoteServer.HTTP_NS = "http://www.w3.org/1999/xx/http#";


/**
 * Literals to parse from RDF.
 */
RemoteServer.REQUIRED_METADATA_LITERALS = [   
    { namespace: RemoteServer.DC_NS, property: "creator"},
    { namespace: RemoteServer.ANNOT_NS, property: "created"},
    { namespace: RemoteServer.DC_NS, property: "date"}, ,
    { namespace: RemoteServer.ANNOT_NS, property: "context"} ];

// type not included b/c it is funky
RemoteServer.REQUIRED_METADATA_RESOURCES = [ { namespace: RemoteServer.ANNOT_NS, property: "annotates"} ];

/**
 * RDF Parsing for Annotation namespace.
 */
RemoteServer.ANNOTATION_TYPE = RemoteServer.ANNOT_NS + "Annotation";

/**
 * RDF Parsing for Reply namespace.
 */
RemoteServer.REPLY_TYPE = RemoteServer.THREAD_NS + "Reply";
					    
RemoteServer.prototype = {
    
    get url() { return this.mUrl; },

    get username() { return this.mUsername; },
    set username(name) { this.mUsername = name; },

    get password() { return this.mPassword; },
    set password(name) { this.mPassword = name; }
};

/**
 * XPCOM QueryInterface function.
 *
 * @param iid
 * @return nsIAnnotationsServer this
 */
RemoteServer.prototype.QueryInterface = function(iid) {
    if (!iid.equals(Components.interfaces.nsIAnnotationsServer) &&
	!iid.equals(Components.interfaces.nsISupports)) {
	    throw Components.results.NS_ERROR_NO_INTERFACE;
	}
    return this;
}

RemoteServer.prototype.init = function(url, username, password) {
    if (this.initialized) {
	throw Components.results.NS_ERROR_ALREADY_INITIALIZED;
    }

    this.mUrl = url;
    this.mUsername = username;
    this.mPassword = password;

    this.initialized = true;
}
    

/**
 * Post a new annotation to the server.  
 *
 * @param data nsIMetadata for a new annotation 
 * @param bodyText string text of annotation body
 * @param encoding string encoding of text
 * @param observer 
 */
RemoteServer.prototype.postAnnotation = function(data, bodyText, encoding, observer) {
  
    var rdfString;  
    try {
      rdfString = RemoteServer.generatePostRDF(data, bodyText);
    }
    catch (e) {
      observer.postAnnotationOnFailure("Insufficient data to post an annotation: " + e);
      return;
    }
    
    var headers = [];
    if (RemoteServer.isUsernameAndPasswordValid(this.mUsername, this.mPassword)) {
	headers.push( ['Authorization', 'Basic ' + RemoteServer.encodeBase64(this.mUsername+":"+this.mPassword) ] );
    }

    // encoding
    if (encoding && encoding != "") {
	headers.push( [ 'Content-Type', 'application/xml; charset=' + encoding ] );
    }
    else {
	headers.push( [ 'Content-Type', 'application/xml' ] );
    }
    
    new RemoteTransmission(RemoteServer.POST, data,
			   observer, headers, 
			   this.mUrl, rdfString, 
			   this.mUrl, null); 
}

/**
 * Load the annotations on this server for the given URI.
 *
 * @param uri string URI to load annotations for
 * @param observer 
 */
RemoteServer.prototype.loadAnnotations = function(uri, observer) {
    var headers = [];
    if (RemoteServer.isUsernameAndPasswordValid(this.mUsername, this.mPassword)) {
	headers.push( [ 'Authorization', 'Basic ' + RemoteServer.encodeBase64(this.mUsername+":"+this.mPassword) ] );
    }
    headers.push( [ 'Accept', 'text/rdf'  ] );

    var openStr = this.mUrl + "?w3c_annotates=" + RemoteServer.escapePlus(uri);
    var sendStr = undefined;

    new RemoteTransmission(RemoteServer.LOAD, uri,
			   observer, headers, 
			   openStr, sendStr, 
			   this.mUrl, null);    
}

/**
 * Load the annotations on this server for the given URI.
 * Called by autoLoad.  Keeps the RemoteTransmission around
 * in case we need to cancel it.
 *
 * @param uri string URI to load annotations for
 * @param observers nsISimpleEnumerator of nsIAnnotationsObserver
 * @param win nsIDOMWindow to pass back to observers
 */
RemoteServer.prototype.autoLoadAnnotations = function(uri, observers, win) {
    var headers = [];
    if (RemoteServer.isUsernameAndPasswordValid(this.mUsername, this.mPassword)) {
	headers.push( [ 'Authorization', 'Basic ' + RemoteServer.encodeBase64(this.mUsername+":"+this.mPassword) ] );
    }
    headers.push( [ 'Accept', 'text/rdf'  ] );

    var openStr = this.mUrl + "?w3c_annotates=" + RemoteServer.escapePlus(uri);
    var sendStr = undefined;

    // passing in multiple observers in the observer slot is sort of 
    // a hack
    new RemoteTransmission(RemoteServer.AUTO_LOAD, uri,
			   observers, headers, 
			   openStr, sendStr, 
			   this.mUrl, win);    
}

/**
 * Delete the annotation at the specified uri.
 *
 * @param uri string uri to delete
 * @param observer 
 */
RemoteServer.prototype.deleteAnnotation = function(uri, observer) {
    var headers = [];
    if (RemoteServer.isUsernameAndPasswordValid(this.mUsername, this.mPassword)) {
	headers.push( [ 'Authorization', 'Basic ' + RemoteServer.encodeBase64(this.mUsername+":"+this.mPassword) ] );
    }
    
    var sendStr = "";

    new RemoteTransmission(RemoteServer.DELETE, uri,
			   observer, headers, 
			   uri, sendStr, 
			   this.mUrl, null); 
}

/**
 * Update the annotation passed in on the server.  
 *
 * @param annotation nsIAnnotation
 * @param observer nsIAnnotationsServerObserver
 */
RemoteServer.prototype.updateAnnotation = function(annotation, observer) {
    var rdfString;  
    try {
	rdfString = RemoteServer.generateUpdateRDF(annotation);
    }
    catch (e) {
      if (annotation.annotationURI) {
	return observer.updateAnnotationOnFailure(annotation.annotationURI,
						  "Insufficient data to update annotation: " + e);
      }
      else {
	return observer.updateAnnotationOnFailure("", "Insufficient data to update annotation: " + e);
      }
    }
    
    var headers = [];
    if (RemoteServer.isUsernameAndPasswordValid(this.mUsername, this.mPassword)) {
	headers.push( [ 'Authorization', 'Basic ' + RemoteServer.encodeBase64(this.mUsername+":"+this.mPassword) ] );
    }

    // encoding
    var encoding = annotation.encoding;
    if (encoding && encoding != "") {
	headers.push( [ 'Content-Type', 'application/xml; charset=' + encoding ] );
    }
    else {
	headers.push( [ 'Content-Type', 'application/xml' ] );
    }
    
    var openStr = this.mUrl + "?replace_source=" + RemoteServer.escapePlus(annotation.annotationURI);

    new RemoteTransmission(RemoteServer.UPDATE, annotation,
			   observer, headers, 
			   openStr , rdfString, 
			   this.mUrl, null); 
}

/**
 * Load the replies for an annotation.
 *
 * @param annotation nsIAnnotation
 * @param observer nsIAnnotationsServerObserver
 */
RemoteServer.prototype.getReplies = function(annotation, observer) {
    if (! annotation.annotationURI) {
	return observer.getRepliesOnFailure(annotation, "Annotation passed into get replies has no URI.");
    }

    var headers = [];
    if (RemoteServer.isUsernameAndPasswordValid(this.mUsername, this.mPassword)) {
	headers.push( [ 'Authorization', 'Basic ' + RemoteServer.encodeBase64(this.mUsername+":"+this.mPassword) ] );
    }
    headers.push( [ 'Accept', 'text/rdf'  ] );

    var openStr =  this.mUrl + "?w3c_replyTree=" + RemoteServer.escapePlus(annotation.annotationURI);

    new RemoteTransmission(RemoteServer.GET_REPLIES, annotation,
			   observer, headers, 
			   openStr, undefined, 
			   this.mUrl, null);   
}

/**
 * Reply to an annotation.
 *
 * @param original nsIAnnotation original annotation
 * @param reply nsIMetadata reply data
 * @param observer nsIAnnotationsServerObserver observer
 */
RemoteServer.prototype.replyToAnnotation = function(original, reply, observer) {
    return observer.replyToAnnotationOnFailure(original, "Sorry, replyToAnnotation is not " +
					       "yet implemented.  As soon as the W3C releases " +
					       "the protocols, it will be switched on.");
}

/**
 * Load the body of an annotation into the "body" Metadata
 * field of the annotation.
 *
 * @param annotation nsIAnnotation to load body for
 * @param observer nsIAnnotationsObserver
 */
RemoteServer.prototype.loadAnnotationBody = function(annotation, observer) {
  
  if ( (! annotation.bodyURI) || annotation.bodyURI == "") {
    observer.loadAnnotationBodyOnFailure(annotation, "The bodyURI field is not defined for this annotation");
    return;
  }

  var headers = [];
  if (RemoteServer.isUsernameAndPasswordValid(this.mUsername, this.mPassword)) {
      headers.push( [ 'Authorization', 'Basic ' + RemoteServer.encodeBase64(this.mUsername+":"+this.mPassword) ] );
  }
  headers.push( [ 'Accept', '*,gzip'  ] );

  var openStr = annotation.bodyURI;
  var sendStr = undefined;

  new RemoteTransmission(RemoteServer.LOAD_BODY, annotation,
			 observer, headers, 
			 openStr, sendStr, 
			 this.mUrl, null);    
}



/**********************
 * PRIVATE FUNCTIONS  *
 **********************/

/**
 * Returns whether a username and password are
 * both valid (ie, exist).
 */
RemoteServer.isUsernameAndPasswordValid = function(username, password) {
    return (username && (username != ""));
}

/**
 * Given an rdfDataSource and the resources and literals to extract,
 * creates <code>Annotation<code>s with these resources and literals in the metadata
 * of the Annotation.  
 *
 * @param rdfDataSrc nsIRDFDataSource to mine for annotations
 * @param resourceNames names of fields of this object will be parsed as resources from the RDF
 * @param literalNames names of fields of this object will be parsed as literals from the RDF
 * @param rdfType one of either ANNOTATION_TYPE or REPLY_TYPE
 * @return Annotation[] of Annotations
 */
RemoteServer.buildAnnotationsFromDataSource = function(rdfDataSrc, rdfType, serverName, encoding) 
{
    return RemoteServer.buildAnnotationsFromResourceEnum(rdfDataSrc, rdfDataSrc.GetAllResources(), rdfType, serverName, encoding);
}

/**
 * Given an rdfDataSource, a resourceEnum (subset of the dataSource) and the resources and literals to extract,
 * creates <code>Annotation<code>s with these resources and literals in the metadata
 * of the Annotation.  
 *
 * @param rdfDataSrc nsIRDFDataSource to mine for annotations
 * @param resourceEnum nsISimpleEnumerator subset of resources to build annotations for
 * @param resourceNames names of fields of this object will be parsed as resources from the RDF
 * @param literalNames names of fields of this object will be parsed as literals from the RDF
 * @param rdfType one of either ANNOTATION_TYPE or REPLY_TYPE
 * @return Annotation[] of Annotations
 */
RemoteServer.buildAnnotationsFromResourceEnum = function(rdfDataSrc, resourceEnum, rdfType, serverName, encoding) 
{
    var annotations = [];

    while(resourceEnum.hasMoreElements()) {

	// get resource
        var res = resourceEnum.getNext();
        res = res.QueryInterface(Components.interfaces.nsIRDFResource);

	// Only consider resources that are annotations
	var typeEnum = rdfDataSrc.GetTargets (res, gRDFService.GetResource(RemoteServer.RDF_NS + "type"), true);
	if (! typeEnum.hasMoreElements()) {
	    continue;
	}

	var annotURI = res.Value;
	var bodyURI  = RemoteServer.getRDFResource(res, RemoteServer.ANNOT_NS, "body", rdfDataSrc);
	
	var annotation = Components.classes["@mozilla.org/annotation;1"].createInstance(Components.interfaces.nsIAnnotation);
	annotation.initialize(serverName, annotURI, bodyURI, encoding);

	// view all the arcs out
	var arcs = rdfDataSrc.ArcLabelsOut(res);
	while (arcs.hasMoreElements()) {
	    var arc = arcs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
	    if (arc.Value == RemoteServer.ANNOT_NS + "body") {
		continue;
	    }
	    else if (arc.Value == RemoteServer.RDF_NS + "type") {
		// if we've already set the type, continue
		if (! (annotation.data.getProperty(RemoteServer.RDF_NS, "type")) ) {

		    // set the initial assumption to Annotation
		    annotation.data.setResourceProperty(RemoteServer.RDF_NS, "type", "Annotation");
		    // sort through the multiple types
		    while (typeEnum.hasMoreElements()) {
			var type = typeEnum.getNext();
			type = type.QueryInterface(Components.interfaces.nsIRDFResource);
			type = type.Value;
			// Delete up to and including the hash,
			// to get a readable form of the type
			var splitType = RemoteTransmission.splitResourceString(type);
			if (splitType.property != 'Annotation') {
			    annotation.data.setResourceProperty(RemoteServer.RDF_NS, "type", type);
			    break;
			}
		    }
		}

		// always continue
		continue;
	    }

	    var splitValue = RemoteTransmission.splitResourceString(arc.Value);

	    // try to get it as a resource
	    var value = RemoteServer.getRDFResource(res, splitValue.namespace, splitValue.property, rdfDataSrc);
	    // if not a resource, try a literal
	    if (! value) {
		value = RemoteServer.getRDFLiteral(res, splitValue.namespace, splitValue.property, rdfDataSrc);
		if (! value) {
		    // it's blank, store it as an empty string literal
		    value = "";
		}
		annotation.data.setLiteralProperty(splitValue.namespace, splitValue.property, value);
	    }
	    else {
		annotation.data.setResourceProperty(splitValue.namespace, splitValue.property, value);
	    }
	}

	annotations.push(annotation);
    }

    return annotations;
}


/**
 * Construct a standard RDF string 
 * to pass to an annotation server 
 *
 * @param meta Metadata for Annotation
 * @param bodyText string text of annotation body
 * @return RDF post string
*/
RemoteServer.generatePostRDF = function( meta, bodyText ) {
    var literals = {};
    var resources = {};
    var namespaces = {};
    var namespacePrefix = "a";
    var namespaceNextNum = 0;

    // setup required namespaces
    namespaces[RemoteServer.ANNOT_NS] = "a";
    namespaces[RemoteServer.RDF_NS] = "r";
    namespaces[RemoteServer.DC_NS] = "dc";
    namespaces[RemoteServer.HTTP_NS] = "h";

    // check for required literals
    for (var i = 0; i < RemoteServer.REQUIRED_METADATA_LITERALS; i++) {
	var literal = RemoteServer.REQUIRED_METADATA_LITERALS[i];
	var aValue = meta.getProperty(literal.namespace, literal.property);

	if (aValue) {
	    literals[literal.namespace + literal.property] = { prefix: namespaces[literal.namespace], property: literal.property, value: aValue };   
	}
	else {
	    throw Components.Exception("Metadata is missing required literal " + 
				       literal.namespace + literal.property + "\n");
	}
    }
    
    // check for required resources
    for (var i = 0; i < RemoteServer.REQUIRED_METADATA_RESOURCES; i++) {
	var resource = RemoteServer.REQUIRED_METADATA_RESOURCES[i];
	var aValue = meta.getProperty(resource.namespace, resource.property);

	if (aValue) {
	    resources[resource.namespace + resource.property] = 
	    { prefix: namespaces[resource.namespace], property: resource.property, value: aValue };
	    
	}
	else {
	    throw Components.Exception("Metadata is missing required resource " + 
				       resource.namespace + resource.property + "\n");
	}
    }

    // check for type, which cannot live in resources b/c of parsing issues
    var type = meta.getProperty(RemoteServer.RDF_NS, "type");
    if (! type) {
	throw Components.Exception("Metadata is missing required resource " +
				   RemoteServer.RDF_NS + "type");
    }
    
 // enumerate the metadata to add additional items
    var properties = {};
    var count = {};
    meta.getAllProperties(count, properties);
    for (var i = 0; i < properties.value.length; i++) {
	var datum = properties.value[i];
	if (! namespaces[datum.namespace] ) {
	    namespaces[datum.namespace] = namespacePrefix + (namespaceNextNum++);
	}
	if (datum.type == datum.LITERAL) {
	    if (! literals[datum.namespace + datum.property]) {
		
		literals[datum.namespace + datum.property] = { prefix: namespaces[datum.namespace], property: datum.property, value: datum.value };
	    }
	}
	else {
	    if (! resources[datum.namespace + datum.property]) {
		resources[datum.namespace + datum.property] = { prefix: namespaces[datum.namespace], property: datum.property, value: datum.value };
		
	    }
	}
    }
    
    var outputArray = [];
    outputArray.push('<?xml version=\"1.0\"?>\n');
    outputArray.push('<r:RDF');
    
    // add namespaces
    for (fullName in namespaces) {
	outputArray.push('\nxmlns:' + namespaces[fullName] + '="' + fullName + '"');
    }

    // close opening RDF tag
    outputArray.push('>\n');

    // add the types
    outputArray.push('<r:Description>\n' +
		     '<r:type resource="http://www.w3.org/2000/10/annotation-ns#Annotation"/>\n');

    // add the resources
    for (prop in resources) {
	var resource = resources[prop];
	if (resource.prefix == "r") {
	    outputArray.push('<' + resource.prefix + ':' + resource.property + ' resource="' +
			     resource.value + '"/>\n');
	}
	else {
	    outputArray.push('<' + resource.prefix + ':' + resource.property + ' r:resource="' +
			     resource.value + '"/>\n');
	}
    }

    // add the literals 
    for (prop in literals) {
	var literal = literals[prop];
	outputArray.push('<' + literal.prefix + ':' + literal.property + '>' +
			 literal.value + '</' + literal.prefix + ':' + literal.property + '>\n');
    }

    // add the body stuff and closing description and RDF tags
    outputArray.push(
	'<a:body>\n<h:Message>\n<h:ContentType>text/html</h:ContentType>\n' +
	//	'<h:ContentLength>' + bodyText.length + '</h:ContentLength>' +
	'<h:Body r:parseType=\"Literal\">' + bodyText +
	'</h:Body>\n</h:Message>\n</a:body>\n' +
	'</r:Description>\n</r:RDF>');

    return outputArray.join("");
}

/**
 * Construct an RDF string to send to the
 * server to update an extant annotation.
 *
 * @param annot extant Annotation
 * @return RDF update string
 */
RemoteServer.generateUpdateRDF = function(annot) {
    var literals = {};
    var resources = {};
    var namespaces = {};
    var namespacePrefix = "a";
    var namespaceNextNum = 0;

    // setup required namespaces
    namespaces[RemoteServer.ANNOT_NS] = "a";
    namespaces[RemoteServer.RDF_NS] = "r";
    namespaces[RemoteServer.DC_NS] = "dc";
    namespaces[RemoteServer.HTTP_NS] = "h";

    // check for required annotation URI
    if ( (!annot.annotationURI) || (annot.annotationURI=="") ) {
      throw Components.Exception("Annotation missing annotationURI");
    }

    // check for required literals
    for (var i = 0; i < RemoteServer.REQUIRED_METADATA_LITERALS; i++) {
	var literal = RemoteServer.REQUIRED_METADATA_LITERALS[i];
	var aValue = annot.data.getProperty(literal.namespace, literal.property);

	if (aValue) {
	    literals[literal.namespace + literal.property] = { prefix: namespaces[literal.namespace], property: literal.property, value: aValue };  
	}
	else {
	    throw Components.Exception("Metadata is missing required literal " + 
				       literal.namespace + literal.property + "\n");
	}
    }
    
    // check for required resources
    for (var i = 0; i < RemoteServer.REQUIRED_METADATA_RESOURCES; i++) {
	var resource = RemoteServer.REQUIRED_METADATA_RESOURCES[i];
	var aValue = annot.data.getProperty(resource.namespace, resource.property);

	if (aValue) {
	    resources[resource.namespace + resource.property] = 
	    { prefix: namespaces[resource.namespace], property: resource.property, value: aValue };
	    
	}
	else {
	    throw Components.Exception("Metadata is missing required resource " + 
				       resource.namespace + resource.property + "\n");
	}
    }

    // check for type, which cannot live in resources b/c of parsing issues
    var type = annot.data.getProperty(RemoteServer.RDF_NS, "type");
    if (! type) {
      throw Components.Exception("Metadata is missing required resource " +
				 RemoteServer.RDF_NS + "type");
    }
    

    // enumerate the metadata to add additional items
    var count = {};
    var properties = {};
    annot.data.getAllProperties(count, properties);
    for (var i = 0; i < properties.value.length; i++) {
	var datum = properties.value[i];
	if (! namespaces[datum.namespace] ) {
	    namespaces[datum.namespace] = namespacePrefix + (namespaceNextNum++);
	}
	if (datum.type == datum.LITERAL) {
	    if (! literals[datum.namespace + datum.property]) {
		
		literals[datum.namespace + datum.property] = { prefix: namespaces[datum.namespace], property: datum.property, value: datum.value };
	    }
	}
	else {
	    if (! resources[datum.namespace + datum.property]) {
		resources[datum.namespace + datum.property] = { prefix: namespaces[datum.namespace], property: datum.property, value: datum.value };
		
	    }
	}
    }
    
    var outputArray = [];
    outputArray.push('<?xml version=\"1.0\"?>\n');
    outputArray.push('<r:RDF');
    
    // add namespaces
    for (fullName in namespaces) {
	outputArray.push('\nxmlns:' + namespaces[fullName] + '="' + fullName + '"');
    }

    // close opening RDF tag
    outputArray.push('>\n');

    // add the types
    outputArray.push('<r:Description about="' + annot.annotationURI + '">\n' +
		     '<r:type resource="http://www.w3.org/2000/10/annotation-ns#Annotation"/>\n');

    // add the resources
    for (prop in resources) {
	var resource = resources[prop];
	// don't add body resource
	if (resource.property == "body") {
	    continue;
	}
	if (resource.prefix == "r") {
	    outputArray.push('<' + resource.prefix + ':' + resource.property + ' resource="' +
			     resource.value + '"/>\n');
	}
	else {
	    outputArray.push('<' + resource.prefix + ':' + resource.property + ' r:resource="' +
			     resource.value + '"/>\n');
	}
    }

    // add the literals 
    for (prop in literals) {
	var literal = literals[prop];
	outputArray.push('<' + literal.prefix + ':' + literal.property + '>' +
			 literal.value + '</' + literal.prefix + ':' + literal.property + '>\n');
    }

   
    // add the body stuff 
    var bodyText = annot.getBodyText();
    if (bodyText) {
	outputArray.push('<a:body>\n<h:Message>\n<h:ContentType>text/html</h:ContentType>\n' +
			// '<h:ContentLength>' + bodyText.length + '</h:ContentLength>' +
			 '<h:Body r:parseType=\"Literal\">' + bodyText +
			 '</h:Body>\n</h:Message>\n</a:body>\n');
    }
    else {
	outputArray.push('<a:body r:resource="' + annot.bodyURI + '"/>\n');
    }

    // close out the Description and RDF
    outputArray.push('</r:Description>\n</r:RDF>');
    	
    return outputArray.join("");
}


// used for base 64 <--> ascii
RemoteServer.ENCODE_ORDER = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/**
 * Transform base 64 string to Ascii string.
 *
 * @param c String in base 64
 * @return String in Ascii
 */
RemoteServer.base64ToAscii = function(c)
{
    return RemoteServer.ENCODE_ORDER.substring(c,c+1);
}



/**
 * Transforms a string from ascii to base 64.
 *
 * @param str ascii string
 * @return String encoded in base 64 
 */
RemoteServer.encodeBase64 = function(str) {
	var result = "";
	var i = 0;
	var sextet = 0;
	var leftovers = 0;
	var octet = 0;

	for (i=0; i < str.length; i++) {
		octet = str.charCodeAt(i);
		switch( i % 3 )
		{
			case 0:
			{
				sextet = ( octet & 0xFC ) >> 2 ;
				leftovers = octet & 0x03 ;

				// sextet contains first character in quadruple
				break;
			}

			case 1:
			{
				sextet = ( leftovers << 4 ) | ( ( octet & 0xF0 ) >> 4 );
				leftovers = octet & 0x0F ;

				// sextet contains 2nd character in quadruple
				break;
			}

			case 2:
			{
				sextet = ( leftovers << 2 ) | ( ( octet & 0xC0 ) >> 6 ) ;
				leftovers = ( octet & 0x3F ) ;

				// sextet contains third character in quadruple
				// leftovers contains fourth character in quadruple
				break;
			}
		}

		result = result + RemoteServer.base64ToAscii(sextet);

		// don't forget about the fourth character if it is there
		if( (i % 3) == 2 )
		{
			result = result + RemoteServer.base64ToAscii(leftovers);
		} 

	}

	// figure out what to do with leftovers and padding
	switch( str.length % 3 )
	{
		case 0:
		{
			// an even multiple of 3, nothing left to do
			break ;
		}
		case 1:
		{
			// one 6-bit chars plus 2 leftover bits
			leftovers =  leftovers << 4 ;
			result = result + RemoteServer.base64ToAscii(leftovers);
			result = result + "==";
			break ;
		}
		case 2:
		{
			// two 6-bit chars plus 4 leftover bits
			leftovers = leftovers << 2 ;
			result = result + RemoteServer.base64ToAscii(leftovers);
			result = result + "=";
			break ;
		}
	}

	return result;
}

/**
 * Does usual url escaping and also
 * escapes + character.
 *
 * @param s String to escape
 * @return escaped String
 */
RemoteServer.escapePlus = function(s) {
  return encodeURI(s).replace(/\+/g, '%2b');
}


/**
 * Get a property from an RDF Literal.
 *
 * @param resource RDF resource
 * @param ns String namespace
 * @param property String property to retrieve
 * @param rdfDataSrc RDF Data Source
 * @return value of literal or undefined if not found
 *  
 */
RemoteServer.getRDFLiteral = function(resource, ns, property, rdfDataSrc) {
    return RemoteServer.getRDFValue (resource, ns, property, rdfDataSrc,
			     Components.interfaces.nsIRDFLiteral);
}

/**
 * Get a property from an RDF resource.
 *
 * @param resource RDF resource
 * @param ns String namespace
 * @param property String property to retrieve
 * @param rdfDataSrc RDF Data Source
 * @return value of property, or undefined if not found
 *  
 */
RemoteServer.getRDFResource = function(resource, ns, property, rdfDataSrc) {
    return RemoteServer.getRDFValue (resource, ns, property, rdfDataSrc,
			     Components.interfaces.nsIRDFResource);
}

/**
 * Get an RDF Value.
 *
 * @param resource RDF resource
 * @param ns String namespace
 * @param property String property to retrieve
 * @param rdfDataSrc RDF Data Source
 * @param iface interface to query
 * @return value of resource or undefined if property not found
 *  
 */
RemoteServer.getRDFValue = function(resource, ns, property, rdfDataSrc, iface) {
    try {
        var RDFNode =
	rdfDataSrc.GetTarget(resource, gRDFService.GetResource(ns + property), true);
        RDFNode = RDFNode.QueryInterface(iface);
        RDFNode = RDFNode.Value;
        return RDFNode;
    } catch (e) {
	//dump("error getting RDF Value: " + e + " for property " + property + "\n");
        return undefined;
    }
}



/**
 * toString.
 *
 * @return string representation of this
 */
RemoteServer.prototype.toString = function() {
    return "[RemoteServer url==" + this.mUrl + "]";
}








/**
 * Wrapper object for an XMLHttpRequest to 
 * contain the necessary additional information
 * (error handling, etc).
 * Need to use this instead of server, because
 * we can create multiple instances of this, whereas
 * two simultaneous posts to the same server could
 * overwrite data.
 *
 * @param operation one of the RemoteServer. operations
 * @param arg arg passed specific to which operation called
 * @param observer nsIAnnotationsObserver
 * @param headers array of 2-element arrays constituting request headers
 * @param openStr String to pass to open
 * @param sendStr String to pass to send
 * @param serverUrl string url of server
 * @param win nsIDOMWindow to pass back to autoload -- optional
 */
function RemoteTransmission(operation, arg, 
			    observer, headers, 
			    openStr, sendStr, 
			    serverUrl, win)
{ 
    this.operation = operation;
    this.arg = arg;
    this.observer = observer;
    this.serverUrl = serverUrl;
    this.win = win;

    this.xhr = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);

    // get a reference named that to embed in closure
    var that = this;
    
    // operation specific details
    switch(operation) {
    case RemoteServer.POST:
	this.xhr.open("POST", openStr);
	// pass in that.arg==metadata for the post operation
	this.xhr.onload = function () { that.finishPost(that.arg) };
	this.xhr.onerror = function () { 
	    that.observer.postAnnotationOnFailure("An error occurred while trying to post an annotation to " + that.serverUrl);
	};	
					
	break;
    
    case RemoteServer.AUTO_LOAD:
	this.xhr.open("GET", openStr);
	this.xhr.onload = function () { that.finishAutoLoad(); };
	this.xhr.onerror = function () { 
	    var observers = that.observer;
	    while(observers.hasMoreElements()) {
		var current = observers.getNext().QueryInterface(Components.interfaces.nsIAnnotationsObserver);
		current.autoLoadFailureForServer(that.serverUrl, that.arg, "Unable to load annotations from server.");
	    }
	};		      
	break;


    case RemoteServer.LOAD:
	this.xhr.open("GET", openStr);
	this.xhr.onload = function () { that.finishLoad(); };
	this.xhr.onerror = function () { 
	    that.observer.loadAnnotationsOnFailure(that.serverUrl, that.arg, "Unable to load annotations from server.");
	};		      
	break;

    case RemoteServer.DELETE:
	this.xhr.open("DELETE", openStr);
	this.xhr.onload  = function () { that.finishDelete(that.arg); };
	this.xhr.onerror = function () { 
	    that.observer.deleteAnnotationOnFailure(that.arg, "An error occurred while trying to delete an annotation " +
						    "from server" + that.serverUrl);
	};
	break;

    case RemoteServer.UPDATE:
	this.xhr.open("POST", openStr);
	this.xhr.onload = function () { that.finishUpdate(that.arg); };
	this.xhr.onerror = function () { 
	    that.observer.updateAnnotationOnFailure(that.arg.annotationURI,
						    "An error occurred while trying to update annotation " +
						    "on server " + that.serverUrl);
	};	
	break;
	
    case RemoteServer.GET_REPLIES:
	this.xhr.open("GET", openStr);
	this.xhr.onload = function () { that.finishGetReplies(that.arg) };
	this.xhr.onerror = function () { 
	    that.observer.getRepliesOnFailure(that.arg, "An error occurred while trying to get replies " +
					      " to an annotation on server " + that.serverUrl);
	};
	break;

    case RemoteServer.REPLY:
	throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    case RemoteServer.LOAD_BODY:
      this.xhr.open("GET", openStr);
      this.xhr.onload = function () { that.finishLoadBody(that.arg); };
      this.xhr.onerror = function () { 
	that.observer.loadAnnotationBodyOnFailure(that.arg, "An error occurred while trying to " +
						  "load the annotation body at " + that.arg.bodyURI);
      };		      
      break;
      
    default:	
	throw Components.Exception("Operation " + operation + " is invalid.");
    }
    
    // set headers
    for (var i=0; i < headers.length; i++) {
	this.xhr.setRequestHeader(headers[i][0], headers[i][1]);
    }
    
    this.xhr.send(sendStr);    
    
    // signify that we're live
    this.live = true;
}

RemoteTransmission.SERVER_OK_STATUS_CODE = 200;

/**
 * Handles the finishing touches
 * after a response is received
 * from a POST.
 */
RemoteTransmission.prototype.finishPost = function(sentMetadata) 
{
    this.live = false;

    try {
	if (this.xhr.status != 200) {
	    this.observer.postAnnotationOnFailure(this.xhr.statusText);
	    return;
	}
    }
    catch(e) {
	dump("error getting status: " + e + "\n");
    }

    // get the encoding
    var contentType = this.xhr.getResponseHeader('Content-Type');
    var encoding = RemoteTransmission.getCharacterEncoding(contentType);
    
    var dataSource = RemoteTransmission.buildDataSourceForUrl(sentMetadata.getProperty(RemoteServer.ANNOT_NS, "annotates"), 
							      this.xhr.responseText);

    var annotations = RemoteServer.buildAnnotationsFromDataSource(dataSource, RemoteServer.ANNOTATION_TYPE, this.serverUrl, encoding);
    if (annotations.length == 0) {
	return this.observer.postAnnotationOnFailure("Unable to create an annotation on remote server " + this.serverUrl);
    }

    /* since a post returns incomplete information, supplement with 
       the metadata we sent to the server */
    var count = {};
    var properties = {};
    sentMetadata.getAllProperties(count, properties);
    for (var i=0; i < properties.value.length; i++) {
	var elt = properties.value[i];
	if (elt.type == elt.RESOURCE) {
	    annotations[0].data.setResourceProperty(elt.namespace, elt.property, elt.value);
	}
	else {
	    annotations[0].data.setLiteralProperty(elt.namespace, elt.property, elt.value);
	}
    }

    this.observer.postAnnotationOnSuccess(annotations[0]);
    return;
}

RemoteTransmission.prototype.finishLoad = function() 
{
    this.live = false;
    var annotatedURL = this.arg;

    if (this.xhr.status != RemoteTransmission.SERVER_OK_STATUS_CODE) {
	this.observer.loadAnnotationsOnFailure(this.serverUrl, annotatedURL, 
					       "The server gave the response " + this.xhr.status + 
					       ": " + this.xhr.statusText);
    }
    else {
	// get the encoding
	var contentType = this.xhr.getResponseHeader('Content-Type');
	var encoding = RemoteTransmission.getCharacterEncoding(contentType);
	
	var dataSource = RemoteTransmission.buildDataSourceForUrl(annotatedURL, this.xhr.responseText);
	var annotations = RemoteServer.buildAnnotationsFromDataSource(dataSource, RemoteServer.ANNOTATION_TYPE, this.serverUrl, encoding);
	this.observer.loadAnnotationsOnSuccess(this.serverUrl, annotatedURL, 
					       annotations.length, annotations);
    }
}

RemoteTransmission.prototype.finishAutoLoad = function() 
{
    this.live = false;
    var annotatedURL = this.arg;
    var observers = this.observer;

    if (this.xhr.status != RemoteTransmission.SERVER_OK_STATUS_CODE) {
	while(observers.hasMoreElements()) {
	    var current = observers.getNext().QueryInterface(Components.interfaces.nsIAnnotationsObserver);
	    current.autoLoadFailureForServer(this.serverUrl, this.win, annotatedURL, 
					     "The server gave the response " + this.xhr.status + 
					     ": " + this.xhr.statusText);
	}
    }
    else {
	// get the encoding
	var contentType = this.xhr.getResponseHeader('Content-Type');
	var encoding = RemoteTransmission.getCharacterEncoding(contentType);
	
	var dataSource = RemoteTransmission.buildDataSourceForUrl(annotatedURL, this.xhr.responseText);
	var annotations = RemoteServer.buildAnnotationsFromDataSource(dataSource, RemoteServer.ANNOTATION_TYPE, this.serverUrl, encoding);
	
	while(observers.hasMoreElements()) {
	    var current = observers.getNext().QueryInterface(Components.interfaces.nsIAnnotationsObserver);
	    current.autoLoadAnnotationsForServer(this.serverUrl, this.win, annotatedURL, 
						 annotations.length, annotations);
	}
    }
}

RemoteTransmission.prototype.finishDelete = function(url) 
{
    this.live = false;
    
    if (RemoteTransmission.isOKStatus(this.xhr.status)) {
	return this.observer.deleteAnnotationOnSuccess(url);
    }
    else {
	return this.observer.deleteAnnotationOnFailure(url, "The server responded with a " + this.xhr.status + 
						       " when trying to delete the annotation.");
    }
}

RemoteTransmission.prototype.finishUpdate = function(oldAnnot) 
{    
    this.live = false;

    if (! RemoteTransmission.isOKStatus(this.xhr.status)) {
	return this.observer.updateAnnotationOnFailure(oldAnnot.annotationURI,
						       "The server responded with a " + this.xhr.status + 
						       " error code when trying to update the annotation.",
						       this.serverUrl);
    }

    // get the encoding
    var contentType = this.xhr.getResponseHeader('Content-Type');
    var encoding = RemoteTransmission.getCharacterEncoding(contentType);
	
    var dataSource = RemoteTransmission.buildDataSourceForUrl(oldAnnot.annotationURI, this.xhr.responseText);
    var newAnnots = RemoteServer.buildAnnotationsFromDataSource(dataSource, RemoteServer.ANNOTATION_TYPE, this.serverUrl, encoding);

    if (newAnnots.length == 0) {
	this.observer.updateAnnotationOnFailure(oldAnnot.annotationURI, 
						"An error occurred while trying to update an annotation on server " +
						this.serverUrl);
	return;
    }

    /* since an update returns incomplete information, supplement with 
       the posted annotation's metadata */
    var count = {};
    var properties = {};
    oldAnnot.data.getAllProperties(count,properties);
    for(var i = 0; i < properties.value.length; i++) {
	var elt = properties.value[i];
	if (elt.type == elt.RESOURCE) {
	  newAnnots[0].data.setResourceProperty(elt.namespace, elt.property, elt.value);
	}
	else {
	  newAnnots[0].data.setLiteralProperty(elt.namespace, elt.property, elt.value);
	}
    }
    
    this.observer.updateAnnotationOnSuccess(newAnnots[0]);
}    

RemoteTransmission.prototype.finishGetReplies = function(original) 
{
    this.live = false;
    // get the encoding
    var contentType = this.xhr.getResponseHeader('Content-Type');
    var encoding = RemoteTransmission.getCharacterEncoding(contentType);
	
    var dataSource = RemoteTransmission.buildDataSourceForUrl(original.annotationURI, this.xhr.responseText);
    var replies = RemoteServer.buildAnnotationsFromDataSource(dataSource, RemoteServer.ANNOTATION_TYPE, this.serverUrl, encoding);
    return this.observer.getRepliesOnSuccess(original, replies.length, replies);
}

RemoteTransmission.prototype.finishLoadBody = function(annotation) {
    this.live = false;
  
    if (! RemoteTransmission.isOKStatus(this.xhr.status)) {
	return this.observer.loadAnnotationBodyOnFailure(annotation, "Unable to load annotation body.  The server replied: " + this.xhr.status + "\n");
    }
    
    var bodyText = this.xhr.responseText;
    
    // get the encoding
    var contentType = this.xhr.getResponseHeader('Content-Type');
    var encoding = RemoteTransmission.getCharacterEncoding(contentType);
    
    annotation.encoding = encoding;
    annotation.setBodyText(bodyText);

    this.observer.loadAnnotationBodyOnSuccess(annotation, bodyText);
}

RemoteTransmission.prototype.abort = function() {
    if (this.live) {
	this.xhr.abort();
	this.live = false;
    }
}

RemoteTransmission._cloneAnnotations = function(oldAnnots) {
      var newAnnots = [];
      for (var i = 0; i < oldAnnots.length; i++) {
	newAnnots.push(oldAnnots[i].clone());
      }
      return newAnnots;
}


/**
 * Utility function to build an RDF data source, given
 * the url of the source and the rdf text.
 *
 * @param strUrl String url of rdf
 * @param rdfText String rdf text
 * @exception if unable to parse RDF text
 * @return nsIRDFDataSource
 */
RemoteTransmission.buildDataSourceForUrl = function(strUrl, rdfText) {
    
    var ios = Components.classes["@mozilla.org/network/io-service;1"].getService (Components.interfaces.nsIIOService);
    var rdfDataSrc = Components.classes["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"].createInstance(Components.interfaces.nsIRDFDataSource);
    var xmlParser = Components.classes["@mozilla.org/rdf/xml-parser;1"].createInstance (Components.interfaces.nsIRDFXMLParser);
    var uri = ios.newURI (strUrl, null, null);
    
    xmlParser.parseString (rdfDataSrc, uri, rdfText);

    return rdfDataSrc;
}

/**
 * Splits a namespace + property Resource string into
 * namespace and property.
 *
 * @param resourceString string of namespace + property
 * @return object with namespace and property fields
 */
RemoteTransmission.splitResourceString = function(resourceString) {
    var i;
    for (i = (resourceString.length - 1); i >= 0; i--) {
	var charAtI = resourceString.charAt(i);
	if (charAtI == '#' || charAtI == '/' || charAtI == ':') {
	    return { namespace: resourceString.substring(0, i+1),
				    property: resourceString.substring(i+1) };
	}
    }

    // couldn't find it -- return null
    return null;
}

/**
 * Returns whether a status code is of
 * the 200 family.
 *
 * @param status integer status code
 * @return whether status is of 200 family
 */
RemoteTransmission.isOKStatus = function(status) {
    return (Math.floor(status / 100) == 2);
}


RemoteTransmission.getCharacterEncoding = function(contentType) {
    if (! contentType) {
	return null;
    }

    // find charset=
    var strToFind = "charset=";
    var start = contentType.indexOf(strToFind);
    if (start == -1) {
	return null;
    }
    start += strToFind.length;
    var end = contentType.indexOf(";", start);
    if (end == -1) {
	    end = contentType.length;
    }
    return contentType.substring(start, end);
}

/********
 * XPCOM
 * stuff
 ********/

var serverModule = {
    firstTime: true,
    
    /*
     * RegisterSelf is called at registration time (component installation
     * or the only-until-release startup autoregistration) and is responsible
     * for notifying the component manager of all components implemented in
     * this module.  The fileSpec, location and type parameters are mostly
     * opaque, and should be passed on to the registerComponent call
     * unmolested.
     */
    registerSelf: function (compMgr, fileSpec, location, type) {
        if (this.firstTime) {
            this.firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        dump("*** Registering nsRemoteAnnotationsServer\n");
        debug("*** Registering nsRemoteAnnotationsServer\n");
        compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
        compMgr.registerFactoryLocation(NS_REMOTE_ANNOTATIONS_SERVER_CID,
                                        "Remote Annotations Server",
                                        NS_REMOTE_ANNOTATIONS_SERVER_PROG_ID, 
                                        fileSpec,
                                        location,
                                        type);
    },

    /*
     * The GetClassObject method is responsible for producing Factory and
     * SingletonFactory objects (the latter are specialized for services).
     */
    getClassObject: function (compMgr, cid, iid) {
        if (!cid.equals(NS_REMOTE_ANNOTATIONS_SERVER_CID)) {
	    throw Components.results.NS_ERROR_NO_INTERFACE;
	}

        if (!iid.equals(Components.interfaces.nsIFactory)) {
	    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
	}

        return this.serverFactory;
    },

    /* factory object */
    serverFactory: {
        /*
         * Construct an instance of the interface specified by iid, possibly
         * aggregating it with the provided outer.  (If you don't know what
         * aggregation is all about, you don't need to.  It reduces even the
         * mightiest of XPCOM warriors to snivelling cowards.)
         */
        createInstance: function (outer, iid) {
            if (outer != null) {
		throw Components.results.NS_ERROR_NO_AGGREGATION;
	    }

            return (new RemoteServer()).QueryInterface(iid);
        }
    },

    /*
     * The canUnload method signals that the component is about to be unloaded.
     * C++ components can return false to indicate that they don't wish to be
     * unloaded, but the return value from JS components' canUnload is ignored:
     * mark-and-sweep will keep everything around until it's no longer in use,
     * making unconditional ``unload'' safe.
     *
     * You still need to provide a (likely useless) canUnload method, though:
     * it's part of the nsIModule interface contract, and the JS loader _will_
     * call it.
     */
    canUnload: function(compMgr) {
        return true;
    }
};

function NSGetModule(compMgr, fileSpec) {
    return serverModule;
}
